#include <string>
#include <fstream>
#include <bitset>
#include <iostream>

// Save the bitmap to a bmp file
static void SaveBitmapToFile(
    std::byte* pBitmapBits,
    int width,
    int height,
    char bits,
    int pad,
    std::string filename
) {

#define BMP_HEADER_SIZE 14
#define BMP_INFO_SIZE   40

    unsigned img_size = height * ((width * (bits / 8)) + pad);
    unsigned bmp_size = BMP_HEADER_SIZE + BMP_INFO_SIZE + img_size;

#define QUAD(X) \
    static_cast<char>(X),\
    static_cast<char>(X >>  8),\
    static_cast<char>(X >> 16),\
    static_cast<char>(X >> 24)

    const char BMPFileHeader[BMP_HEADER_SIZE] = {
        'B', 'M',
        QUAD(bmp_size),
        0,0,
        0,0,
        QUAD(54)
    };
    const char BMPInfoHeader[BMP_INFO_SIZE] = {
        40,0,0,0, // size
        QUAD(width),  // w
        QUAD(height), // h
        1,0,     // planes
        bits,0,  // bpp
        0,0,0,0, // compression
        QUAD(img_size), // bitmap size
        0,0,0,0, // hres
        0,0,0,0, // vres
        0,0,0,0, // palette size
        0,0,0,0  // important colours
    };

    std::ofstream outFile;
    outFile.open(filename, std::ios::out | std::ios::binary);
    if (!outFile) {
        std::cout << "Failed to open \"" << filename << "\"" << std::endl;
        return;
    }
    outFile.write(BMPFileHeader, BMP_HEADER_SIZE);
    outFile.write(BMPInfoHeader, BMP_INFO_SIZE);
    outFile.write(reinterpret_cast<const char*>(pBitmapBits), img_size);
    outFile.close();
}


static void convert_row (std::byte *buf, int *pos, int **bitplane, int p) {

    static const unsigned char colours[16][3] = { // An array of colour values
        //BBB  GGG  RRR
        { 255, 255, 255 }, /* White    */
        {   0,  48,  64 }, /* Index  1 */
        {   0,  72, 112 }, /* Index  2 */
        {  56, 104, 168 }, /* Index  3 */
        {   0, 152, 216 }, /* Index  4 */
        {   8, 216, 248 }, /* Index  5 */
        { 255, 255, 255 }, /* Index  6 */
        { 208, 208, 208 }, /* Index  7 */
        { 160, 160, 160 }, /* Index  8 */
        { 112, 120, 120 }, /* Index  9 */
        {   0,   0,  72 }, /* Index 10 */
        {   8,   8, 160 }, /* Index 11 */
        {  16,  16, 248 }, /* Index 12 */
        {   8, 104, 248 }, /* Index 13 */
        {   0, 248,   0 }, /* Index 14 */
        {   0,   0,   0 }  /* Index 15 */
    };

    for (int N = 7; N >= 0; N--) {
        std::bitset<4> bs;
        bs.set(0, (bitplane[0][p] >> N) & 1);
        bs.set(1, (bitplane[1][p] >> N) & 1);
        bs.set(2, (bitplane[2][p] >> N) & 1);
        bs.set(3, (bitplane[3][p] >> N) & 1);
        unsigned long Index = bs.to_ulong();
        int c = *pos;
        buf[c++] = (std::byte)colours[Index][0]; // Blue
        buf[c++] = (std::byte)colours[Index][1]; // Green
        buf[c++] = (std::byte)colours[Index][2]; // Red
        *pos = c;
    }
}


int main () {

    // Fix these before attempting to run this!
    std::string File_Path = "";
    std::string ROM_Path  = "dkc.sfc";

    char oData[8000];
    std::ifstream inFile;
    inFile.open(ROM_Path);
    if (!inFile) {
        std::cout << "Failed to open \"" << ROM_Path << "\"" << std::endl;
        return 1;
    }
    inFile.seekg(0x347519 + 10);
    inFile.read(oData, 8000);
    if (!inFile.good()) {
        std::cout << "Failed to read data from input file." << std::endl;
        return 1;
    }
    inFile.close();

    int E = 0;

    // there are only 26 letters, so we have 26 iterations
    for (int X = 0; X <= 25; X++) {

        // These allocations are kept the same
        // but there are a few ways to make them a bit simpler.

        // Letter K -- Top Left Tile
        int* Bitplane0_Top_Left_ROW = new int[8];
        int* Bitplane1_Top_Left_ROW = new int[8];
        int* Bitplane2_Top_Left_ROW = new int[8];
        int* Bitplane3_Top_Left_ROW = new int[8];

        // Letter K -- Top Right Tile
        int* Bitplane0_Top_Right_ROW = new int[8];
        int* Bitplane1_Top_Right_ROW = new int[8];
        int* Bitplane2_Top_Right_ROW = new int[8];
        int* Bitplane3_Top_Right_ROW = new int[8];

        // Letter K -- Bottom Left Tile
        int* Bitplane0_Bottom_Left_ROW = new int[8];
        int* Bitplane1_Bottom_Left_ROW = new int[8];
        int* Bitplane2_Bottom_Left_ROW = new int[8];
        int* Bitplane3_Bottom_Left_ROW = new int[8];

        // Letter K -- Bottom Right Tile
        int* Bitplane0_Bottom_Right_ROW = new int[8];
        int* Bitplane1_Bottom_Right_ROW = new int[8];
        int* Bitplane2_Bottom_Right_ROW = new int[8];
        int* Bitplane3_Bottom_Right_ROW = new int[8];

        // the assignments can be reduced to a pattern,
        // and iterated with a loop.
        // Can they be reduced further?
        for (int i = 0; i < 8; i++) {
            int offset = E + 2*(7-i);
            Bitplane0_Top_Left_ROW[i]     = oData[offset];
            Bitplane1_Top_Left_ROW[i]     = oData[offset +   1];
            Bitplane2_Top_Left_ROW[i]     = oData[offset +  16];
            Bitplane3_Top_Left_ROW[i]     = oData[offset +  17];
            Bitplane0_Top_Right_ROW[i]    = oData[offset +  32];
            Bitplane1_Top_Right_ROW[i]    = oData[offset +  33];
            Bitplane2_Top_Right_ROW[i]    = oData[offset +  48];
            Bitplane3_Top_Right_ROW[i]    = oData[offset +  49];
            Bitplane0_Bottom_Left_ROW[i]  = oData[offset +  64];
            Bitplane1_Bottom_Left_ROW[i]  = oData[offset +  65];
            Bitplane2_Bottom_Left_ROW[i]  = oData[offset +  80];
            Bitplane3_Bottom_Left_ROW[i]  = oData[offset +  81];
            Bitplane0_Bottom_Right_ROW[i] = oData[offset +  96];
            Bitplane1_Bottom_Right_ROW[i] = oData[offset +  97];
            Bitplane2_Bottom_Right_ROW[i] = oData[offset + 112];
            Bitplane3_Bottom_Right_ROW[i] = oData[offset + 113];
        }

        // E is set here for the next iteration
        E = E + 128 + 10;

        // 16 wide, 16 tall, 3 bytes per pixel = 16*16*3 = 768
        std::byte* buf = new std::byte[768];
        int c = 0;

        // If we group the bitmap arrays together,
        // it makes it a bit easier to pass them to a function
        int *bitplane[][4] = {
            { Bitplane0_Top_Left_ROW,     Bitplane1_Top_Left_ROW,     Bitplane2_Top_Left_ROW,     Bitplane3_Top_Left_ROW     },
            { Bitplane0_Top_Right_ROW,    Bitplane1_Top_Right_ROW,    Bitplane2_Top_Right_ROW,    Bitplane3_Top_Right_ROW    },
            { Bitplane0_Bottom_Left_ROW,  Bitplane1_Bottom_Left_ROW,  Bitplane2_Bottom_Left_ROW,  Bitplane3_Bottom_Left_ROW  },
            { Bitplane0_Bottom_Right_ROW, Bitplane1_Bottom_Right_ROW, Bitplane2_Bottom_Right_ROW, Bitplane3_Bottom_Right_ROW }
        };

        // convert each tile one row at a time
        for (int p = 0; p < 8; p++) {
            convert_row(buf, &c, bitplane[2], p); // Bottom Left
            convert_row(buf, &c, bitplane[3], p); // Bottom Right
        }

        for (int p = 0; p < 8; p++) {
            convert_row(buf, &c, bitplane[0], p); // Top Left
            convert_row(buf, &c, bitplane[1], p); // Top Right
        }


        // Generate a filename and create the bitmap
        std::string filename = File_Path // The destination path
                             + std::to_string(X) // The iteration number
                             + "_Letter_"
                             + static_cast<char>(X+0x41) // convert number to ASCII
                             + ".bmp";
        SaveBitmapToFile(buf, 16, 16, 24, 0, filename);


        // Delete allocations
        delete[] buf;
        delete[] Bitplane0_Top_Left_ROW;
        delete[] Bitplane1_Top_Left_ROW;
        delete[] Bitplane2_Top_Left_ROW;
        delete[] Bitplane3_Top_Left_ROW;
        delete[] Bitplane0_Top_Right_ROW;
        delete[] Bitplane1_Top_Right_ROW;
        delete[] Bitplane2_Top_Right_ROW;
        delete[] Bitplane3_Top_Right_ROW;
        delete[] Bitplane0_Bottom_Left_ROW;
        delete[] Bitplane1_Bottom_Left_ROW;
        delete[] Bitplane2_Bottom_Left_ROW;
        delete[] Bitplane3_Bottom_Left_ROW;
        delete[] Bitplane0_Bottom_Right_ROW;
        delete[] Bitplane1_Bottom_Right_ROW;
        delete[] Bitplane2_Bottom_Right_ROW;
        delete[] Bitplane3_Bottom_Right_ROW;
    }

    return 0;
}

